// gates.cpp
#include "stdafx.h"
#include <math.h>
#include "gates.h"

const string Inverter::code = "INV";
const string Buffer::code = "BUF";
const string And::code = "AND";
const string Nand::code = "NAND";
const string Or::code = "OR";
const string Nor::code = "NOR";
const string Xor::code = "XOR";
const string Xnor::code = "XNOR";

eDevice *eCircuit::doDigital(string type)
{
	if (type==Inverter::code) return new Inverter();
	if (type==Buffer::code) return new Buffer();
	if (type==And::code) return new And();
	if (type==Nand::code) return new Nand();
	if (type==Or::code) return new Or();
	if (type==Nor::code) return new Nor();
	if (type==Xor::code) return new Xor();
	if (type==Xnor::code) return new Xnor();
	return 0;
}

int Inverter::eval(eNode *nodes[])
{
	eNode *inp = nodes[ndx[0]];
	eNode *oup = nodes[ndx[1]];
	if (!inp->isDefined()) return 1;
	double value = 0.5;
	if (inp->value<eNode::low) value = 1.0;
	else if (inp->value>eNode::high) value = 0.0;
	oup->set(value);
	return 0;
}

void eCircuit::reset_nodes()
{
	for (int i=0; i<ncount; i++) {
		eNode *node = nodes[i];
		node->defined = false;
	}
}

int eCircuit::eval()
{
	int count = 0;
	eDevice *dev;
	for (int i=0; i<ndevices; i++) {
		dev = devices[i];
		count += dev->eval(nodes);
	}
	return count;
}

int eCircuit::run()
{
	int count, prev_count;
	
	prev_count = count = eval();
	cycles=1;
	while (count) {
		count = eval();
		cycles++;
		if (count == prev_count) break;
		prev_count = count;
	}
	return count;
}


int aoi(eNode *arg[3], int in, int out)
{
	bool defined = true;
	double result = (out? 0: 1);
	eNode *node;
	for (int i=1; i<3; i++) {
		node = arg[i];
		if (!node->isDefined()) defined = false;
		else if (fabs(node->value-in)<0.2) {
			result = out;
			defined = true;
			break;
		}
	}
	if (!defined) return 1;
	arg[0]->set(result);
	return 0;
}

int And::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	return aoi(arg,0,0);
}

int Nand::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	return aoi(arg,0,1);
}
int Or::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	return aoi(arg,1,1);
}
int Nor::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	return aoi(arg,1,0);
}

int Xor::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	if (!arg[1]->isDefined() || !arg[2]->isDefined()) return 1;
	if (arg[1]->isLow() && arg[2]->isLow()) arg[0]->set(0.0);
	else if (arg[1]->isHigh() && arg[2]->isHigh()) arg[0]->set(0.0);
	else arg[0]->set(1.0);
	return 0;
}
int Xnor::eval(eNode *nodes[])
{
	eNode *arg[3];
	for (int i=0; i<3; i++) arg[i] = nodes[ndx[i]];
	if (!arg[1]->isDefined() || !arg[2]->isDefined()) return 1;
	if (arg[1]->isLow() && arg[2]->isLow()) arg[0]->set(1.0);
	else if (arg[1]->isHigh() && arg[2]->isHigh()) arg[0]->set(1.0);
	else arg[0]->set(0.0);
	return 0;
}
#if defined(EXTRA)
void eDevice::reset()
{
	reset_nodes();
	POSITION pos;
	eDevice *dev;
	pos = devices.GetHeadPosition();
	while (pos != NULL) {
		dev = devices.GetNext(pos);
		dev->reset();
	}
}






/*
* two-input device
*/

eDevice & eDevice::Create(eNode &A, eNode &B, eNode &Out)
{
	add(A);
	add(B);
	add(Out);
	return *this;
}

/*
*  three-input device
*/

eDevice & eDevice::Create(eNode &A, eNode &B, eNode &C, eNode &Out)
{
	add(A);
	add(B);
	add(C);
	add(Out);
	return *this;
}

/*
*   n-input device
*/

eDevice & eDevice::Create(eNode *inp, int n, eNode &Out)
{
	int i;
	for (i=0; i<n; i++) add(inp[i]);
	add(Out);
	return *this;
}


/*
*  Inverter (one input)
*/

INVERT & INVERT::Create(eNode &in, eNode &out)
{
	add(in);
	add(out);
	return *this;
}




/*
*  Buffer (one input)
*/

BUFFER & BUFFER::Create(eNode &in, eNode &out)
{
	add(in);
	add(out);
	return *this;
}

int BUFFER::eval()
{
	eNode *head, *tail;
	head = nodes.GetHead();
	tail = nodes.GetTail();
	ASSERT(head);
	ASSERT(tail);
	if (!head->isDefined()) return 1;
	tail->setValue(head->value? 1: 0);
	return 0;
}

/*
*  XOR gate
*/

int XOR::eval()
{
	int i;
	POSITION pos, tail;
	eNode *node;
	int values[2];
	pos = nodes.GetHeadPosition();
	tail = nodes.GetTailPosition();
	for (i=0; i<2; i++) {
		if (pos == tail) return 1; // list too short
		node = nodes.GetNext(pos);
		if (!node->isDefined()) return 1; // input undefined
		values[i] = (node->value? 1: 0);
	}
	nodes.GetTail()->setValue(values[0]^values[1]);
	return 0;
}

#endif